import envrihub
from envrihub.data_access.models import DataAccessObject, OpenAPI3DataAccess
from envrihub.data_access.open_api_client import build_access_class, recursion_limit_handler_void
from prance import ResolvingParser


def test_recursive_spec_parsing():
    parser = ResolvingParser('https://catalogue.staging.envri.eu/api/v1/openapi.json', recursion_limit_handler=recursion_limit_handler_void)
    parser.parse()
    specification = parser.specification
    assert isinstance(specification, dict), f'Parsed specfication is {type(specification)} instead of good ol dict'

def test_openapi_client():
    parser = ResolvingParser('https://catalogue.staging.envri.eu/api/v1/openapi.json', recursion_limit_handler=recursion_limit_handler_void)
    parser.parse()
    specification = parser.specification
    dao = build_access_class(specification, base_address='https://catalogue.staging.envri.eu/')()
    assert isinstance(dao, OpenAPI3DataAccess), 'Wrong DAO type found'

def test_openapi_client_call():
    parser = ResolvingParser('https://catalogue.staging.envri.eu/api/v1/openapi.json', recursion_limit_handler=recursion_limit_handler_void)
    parser.parse()
    specification = parser.specification
    dao = build_access_class(specification, base_address='https://catalogue.staging.envri.eu/')
    dao_instance = dao()
    result = dao_instance.resourcessearch(q='minchia')
    assert isinstance(result, dict), 'OpenAPI client does not manage to make a proper GET request'


import inspect

from .models import RESTfulAPIDataAccess
from .utils import call_restful_api

def _sanitize_base_url(url):
    # the base URL as registered in the COS looks like this:
    # https://www.marinespecies.org/rest/AphiaAttributesByAphiaID/{id}{?include_inherited}
    # so it can have both route and query parameters in it. When you see something like }{ 
    # it is likely route paramters are over
    if '}{' in url:
        return url.split('}{', 1)[0] + '}'
    else:
        return url



def build_epos_catalogue_metadata_access(base_url:str, parameters:list, content_type:str='application/json', description = '', service_name = 'Service')->type:
    '''builds a class with custom access method inheriting from RESTfulAPIDataAccess'''
    # build the access method's docstring
    docstring = description + '\n\nParameters\n----------\n'
    for par in parameters:
        docstring = docstring +  str(par.get('name')) + ': ' + 'default = ' + str(par.get('defaultValue')) + '\n\t' + str(par.get('label')) + '\n'
    # build the access method
    access_method = build_access_function(base_url, params=parameters, response_content_type=content_type, docstring = docstring)
    # and poop out the class
    return type(service_name, tuple([RESTfulAPIDataAccess]), {'access':access_method})

def build_access_function(base_url, params,  response_content_type,
                         docstring):
    ## parameters is a list of objects like this
    '''
    {'defaultValue': 'false',
    'label': 'CategoryID',
    'name': 'include_inherited',
    'required': False,
    'type': 'boolean'}
    '''
    # we need to translate it into a dictionary of parameter name -> default value pairs
    pardict = {par.get('name'):par.get('defaultValue') for par in params}
    # then the base_url can contain a mixture of query and route parameters, we have to fix that
    base_url = _sanitize_base_url(base_url)
    # gogadget closure
    def access(self, **kwargs): # the only way to dynamically build a method's signature is to use **kwargs and then do some deep black magic with inspect
        # build request parameters
        #print(kwargs)
        request_paramaters = {p:(kwargs[p] if (p in kwargs) else v) for p, v in pardict.items()}
        # the following is because of URL parameters
        try:
            query_url = base_url.format(**request_paramaters)
        except:
            query_url = base_url
        # now the ApplicationServerPool object should handle the request
        return call_restful_api(url = query_url, params = request_paramaters, headers = {},
                             method='GET', additional_headers= kwargs.get('additional_headers'),
                                expected_response_content_type = response_content_type)
    # adding docstring
    access.__doc__ = docstring
    # and signature
    signature_params = [inspect.Parameter(name="self", kind = inspect.Parameter.POSITIONAL_ONLY)]
    signature_params = signature_params + [inspect.Parameter(name=p, default = v, kind = inspect.Parameter.KEYWORD_ONLY) for p, v in pardict.items()]
    signature_params.append(inspect.Parameter(name='additional_headers', default = None, kind = inspect.Parameter.KEYWORD_ONLY, annotation=dict))
    #print(signature_params)
    access.__signature__ = inspect.Signature(parameters=signature_params)
    return access



class Resource():
    def __init__(self, id):
        self.id = id
    
class Dataset(Resource):
    '''This class encapsulates datasets in the ENVRI-HUB.
    it is made up of two attributes:
    + Data: the actual data, either a datastructure or a data access object
    + Metadata: information about the data, expressed as a dictionary-like structure
    '''
    def __init__(self, id):
        self.id = id

class Service(Resource):
    pass


from datetime import datetime
from collections.abc import Iterator
from .cos.catalogue import EposCatalogue
import shapely
import logging

from .cos.models import Distribution


class Hub():

    def __init__(self, **kwargs):
        self.catalogue = EposCatalogue(**kwargs)
    
    def search_catalogue(self, query:str = '', start_date:datetime=None, end_date:datetime=None, geography = None, exv = None, provider=None)->Iterator[Distribution]:
        '''
        Parameters
        ----------
        query:str

        start_date:datetime 

        end_date: datetime

        geography: str | shapely.Geometry | tuple
            geographcial bounds in WKT format, with WGS84 coordinates. Altenratively
            also Shapely Geometry objects and tuples can be accepted

        exv:str|list[str]
            essential variables to lookup

        provider: str

        '''
        # parse wkt geogrpahy into a Shape
        if isinstance(geography, str):
            try:
                polygon=shapely.from_wkt(geography )
                bbox = polygon.bounds
            except:
                logging.error('Invalid WKT geography')
        elif isinstance(geography, shapely.Geometry):
            bbox = geography.bounds
        elif isinstance(geography, tuple):
            bbox = geography
        else:bbox = None 
        # here we are not using yield from becasue we may want to add some further
        # processing/reshaping here in the near future...
        for i in self.catalogue.search(query=query, bbox=bbox, startDate=start_date, endDate=end_date, exvs=exv, organisations=provider):
            yield i

    def fetch_from_catalogue(self, resource_id)->Distribution:
        '''Retrieves a specific resource from the catalogue'''
        return self.catalogue.retrieve_resource(resource_id)


'''Methods and classes to access Argo data'''
import argopy
import numpy as np
import time
from datetime import datetime
from argopy import DataFetcher
from SPARQLWrapper import SPARQLWrapper, JSON
from .models import DataAccessObject

def to_argo_datetime(dt:datetime):
    '''converts a Python datetime to the Argo format, i.e. 2007-09-01'''
    return dt.isoformat()

class ArgoError(Exception):
    def __init__(self, message):            
        super().__init__(message)

class InvalidArgoFormatError(ArgoError):
    def __init__(self, message):            
        super().__init__(message)

class ArgoDAO(DataAccessObject):

    def __init__(self, src, ds='phy', mode= 'standard'):
        argopy.set_options(src=src,ds=ds,mode=mode)
        self.__datafetcher = DataFetcher()
        super().__init__()
    
    def set_options(self, **kwargs):
        argopy.set_options(**kwargs)
        # proably we need to re-instantiate the data fetcher class at this point
        self.__datafetcher = DataFetcher()

    def __export(self, f, format, **kwargs):
        if format in ['xarray', 'xr']:
            return f.to_xarray(**kwargs)
        elif format in ['pandas', 'dataframe', 'csv', 'df']:
            return f.to_dataframe(**kwargs)
        elif format in ['dataset', 'netcdf', 'ecmwf']:
            return f.to_dataset(**kwargs)
        else:
            raise InvalidArgoFormatError(f'Format {format} is not recognized by Argopy, try "xarray", "dataframe" or "netcdf".')

    def access(self, method = 'region', **kwargs):
        '''
        Access pethod to the Argo web services

        Parameters
        ----------
        method:str, default:'region'
            The Argopy method you want to use to access your data. It can be either
            'region', 'float', or 'profile'. Check out Argo's documentation for more
            information https://argopy.readthedocs.io/en/latest/api.html#id3

        out_format:str, default:'xarray'
            the output format you want to be returned, it can be either 'xarray',
            'dataframe', or 'netcdf'.

        lon_west:float
            western bounding cube limit, expressed in WGS84 coordinates
        
        lon_east:float
            eastern bounding cube limit, expressed in WGS84 coordinates
        
        lat_south:float
            southern bounding cube limit, expressed in WGS84 coordinates
        
        lat_north:float
            northern bounding cube limit, expressed in WGS84 coordinates
        
        depth_min:float
            bottom bounding cube limit, expressed in meters
        
        depth_max:float
            top bounding cube limit, expressed in WGS84 meters
        
        date_min:datetime
            time interval lower bound
        
        date_max:detetime
            time interval upper bound
        '''
        
        if method in ['region', 'demo']:
            return self.region(**kwargs)
        elif method in ['float']:
            return self.float_access(**kwargs)
    
    def region(self, lon_west,lon_east,lat_south,lat_north,depth_min,depth_max,date_min,date_max, out_format='xarray', **kwargs):
        '''
        Access method to the Argo web services

        Parameters
        ----------
        lon_west:float
            western bounding cube limit, expressed in WGS84 coordinates
        
        lon_east:float
            eastern bounding cube limit, expressed in WGS84 coordinates
        
        lat_south:float
            southern bounding cube limit, expressed in WGS84 coordinates
        
        lat_north:float
            northern bounding cube limit, expressed in WGS84 coordinates
        
        depth_min:float
            bottom bounding cube limit, expressed in meters
        
        depth_max:float
            top bounding cube limit, expressed in WGS84 meters
        
        date_min:datetime
            time interval lower bound
        
        date_max:detetime
            time interval upper bound
        
        out_format:str, default:'xarray'
            the output format you want to be returned, it can be either 'xarray',
            'dataframe', or 'netcdf'.
        ''' 
        f = self.__datafetcher.region([lon_west,lon_east,lat_south,lat_north,depth_min,depth_max,date_min,date_max])
        return self.__export(f, out_format, **kwargs)
    
    def float_access(self, wmo, out_format, **kwargs):
        '''
        Retrieves data from one or more specific float in the Argo monitoring network

        Parameters
        ----------
        wms:float or list
            WMS identifier(s) of the float(s) to load data from

        out_format:str, default:'xarray'
            the output format you want to be returned, it can be either 'xarray',
            'dataframe', or 'netcdf'.
        '''
        f = self.__datafetcher.float(wmo)
        return self.__export(f, out_format, **kwargs)

    


from collections.abc import Iterable
from envrihub.cos.catalogue import EposCatalogue
from envrihub.cos.models import Distribution
from envrihub.data_access.models import DataAccessObject
TEST_CATALOGUE_ADDRESS ='https://catalogue.staging.envri.eu/api/v1'

def test_distribution():
    cos =  EposCatalogue(TEST_CATALOGUE_ADDRESS )
    results = list(cos.search('beacon'))
    dist = cos.retrieve_resource(results[0].id)
    assert isinstance(dist, Distribution), f'{str(dist)} is not a Distribution object'

def test_distribution_dao():
    cos =  EposCatalogue(TEST_CATALOGUE_ADDRESS )
    results = list(cos.search('beacon'))
    dist = cos.retrieve_resource(results[0].id)
    assert dist.id == results[0].id, f'Catalogue returned id {str(dist.id)} which is different from the requested one'

def test_dao_access():
    cos =  EposCatalogue(TEST_CATALOGUE_ADDRESS )
    results = list(cos.search('marine species'))
    dist = cos.retrieve_resource(results[1].id)
    data = dist.dao.access()
    assert isinstance(data, Iterable), 'Unable to parse JSON. Probably connection does not work'

def test_catalogue_search_free_text():
    cos =  EposCatalogue(TEST_CATALOGUE_ADDRESS )
    results = list(cos.search('marine species'))
    assert len(results)>1, 'Catalogue does not find sh*t with free text search'

def test_catalogue_search_coords():
    cos =  EposCatalogue(TEST_CATALOGUE_ADDRESS )
    results = list(cos.search(bbox = (10.70, 36.17, 28.98, 48.34)))
    assert len(results)>1, 'Catalogue does not find sh*t with geo search'

def test_file_access_from_cos():
    cos =  EposCatalogue(TEST_CATALOGUE_ADDRESS )
    results = list(cos.search('beacon'))
    dist = cos.retrieve_resource(results[0].id)
    assert isinstance(dist.dao.access(), bytes), 'File access does not download bytes object'


import inspect


def delegates(to=None, keep=False):
    "Decorator: replace `**kwargs` in signature with params from `to`"
    def _f(f):
        if to is None: to_f,from_f = f.__base__.__init__,f.__init__
        else:          to_f,from_f = to,f
        sig = inspect.signature(from_f)
        sigd = dict(sig.parameters)
        k = sigd.pop('kwargs')
        s2 = {k:v for k,v in inspect.signature(to_f).parameters.items()
              if v.default != inspect.Parameter.empty and k not in sigd}
        sigd.update(s2)
        if keep: sigd['kwargs'] = k
        from_f.__signature__ = sig.replace(parameters=sigd.values())
        return f
    return _f




from io import StringIO
import io
import json
import requests
import logging
from retry import retry
import pandas as pd
from pandas.core.common import standardize_mapping
from pandas.core.dtypes.cast import maybe_box_native # this is because of Scalars in dataframes...

# list of file-like objects for parsing purposes
FILE_LIKE_OBJECTS = [io.TextIOBase, io.BufferedIOBase, io.RawIOBase, io.IOBase]

def _parse_response_format(response:requests.Response, content_type:str):
    '''Helper function that tries to parse exacly one MIME type'''
    #print('Content Type is ' + content_type)
    if content_type in ['json', 'application/json']:
            return response.json()
    elif content_type.startswith('text'):
        if content_type in ['text/csv', 'csv']:
            # special case: we are receiving a CSV string!
            try:
                return pd.read_csv(StringIO(response.text))
            except:
                logging.warning('Not able to parse CSV content, returning text data')
                return response.text
        else: 
                return response.text
    else:
        logging.warning('Unknown content type, returning bytes data')
        return response.content

def parse_response(response:requests.Response, expected_content_type:str|list = 'json'):
    '''
    Tries its best to parse a requests response object into a 
    manageable data structure
    
    Parameters
    ----------
    response: requests.Response
        whatever the call to the Web service returned
    
    expected_content_type: str or list
        the expected content type, it should be a MIME type like 'application/json'
        or 'text/csv'. It's the type the method will try to parse if no explicit
        indication is included in the response headers. You can also pass down a list
        of MIME types and the function will try to parse find which one works. 
        Default value: json
    '''
    # if content type is explicitly declared, we'll try to use it
    declared_content_type = response.headers.get('content-type')
    if declared_content_type:
        content_type = declared_content_type
    else:
        # otherwise we'll use the expected one passed as a parameter
        content_type = expected_content_type
    # if we have a list of possible formats, we'll try to parse them
    # all until we find the one that actually matches the data
    if isinstance(content_type, list):
        for c_t in content_type:
            try:
                return _parse_response_format(response, c_t)
            except:
                # wrong format, let's fail silently then
                logging.warning(f'Response is not a valid {c_t}')
    else:            
        try:
            return _parse_response_format(response, content_type)
        except Exception as e:
            #print(e)
            logging.warning('Not able to parse response content, returning bytes data')
            try:
                return response.content
            except:
                raise IOError('Unable to read server response')

def prepare_payload(payload, content_type='application/json'):
    '''
    Tries its best to prepare an object to be sent as request 
    payload
    
    Parameters
    ----------
    payload: obj 
        whatever object you intend to send
    
    content_type: str
        the expected content type, it should be a MIME type like 'application/json'
        or 'text/csv'. It's the type the method will try to encode. Default value: json
    '''
    if payload is None:
        return None
    # serialize payload in an appropriate way
    if content_type in ['application/json', 'json']:
        # if the user already converted the payload to a JSON string *somehow*
        # we'll just pass that down, otherwise, we'll floss it with the JSON DUMPS
        # method
        if isinstance(payload, str):
            return payload
        return json.dumps(payload)
    elif content_type.startswith('text'):
        return str(payload)
    elif content_type in ['application/octet-stream', 'stream', 'binary']:
        # is the object a string? And if yes, is it a path?
        if isinstance(payload, str):
            # three options: it's a path or  it's already a binary encoded octet stream. Or the user
            # is an asshole and passed down something nonsensical.
            # let's try to open the string as if were a file
            try:
                with open(payload, 'rb') as f:
                    return f.read()
            except:
                # if anything goes badly, then probably the user already
                # dumped that to a bytes string and we'll just send that
                return payload
        # is payload a file-like object?
        if any([isinstance(payload, t) for t in FILE_LIKE_OBJECTS]):
            return payload.read()
    # if everything fails, we just send out an empty object.
    # yes, so it'll fail silently. Is it evil? Indeed... but
    # we revel in reseachers' tears.
    return None

def call_restful_api(url, data:object=None, params:dict = {}, path_params ={}, headers:dict = {'Content-Type': 'application/json'}, files = None,
                     method:str='GET', additional_headers:dict= {}, payload_content_type:str ='application/json',
                     expected_response_content_type:str = 'application/json'):
    '''
    Calls a RESTful API
    Parameters
    ----------
    
    url: str 
        the URL of the API method to invoke, it can include parameters in curly braces, like
        /function/{par1}/{par2}
    data: object 
        an object to be sent as payload
    params:dict    
        a dictionary of parameters to be sent as query parameters
    path_params:dict
        if your URL has parameters in it, put the values in this dictionary
    headers:dict
        the key-value pairs to be sent as headers
    files: object
        an object to be sent as a multipart attachement
    method:str, default:'GET'
        the REST method to invoke
    additional_headers:dict, default:{}
        more header parameters. This dictionary will be merged with the *headers* one.
        what's the use of this? Well, let's say that some methods in this package really
        need it becasue of closures and decorators, but if you are not into such black
        magic, probably you'll never need to use this parameter.
    payload_content_type:str, default:'application/json'
        the type of content you are going to send, if any. It's how the payload will be serialized.
    expected_response_content_type:str default:'application/json'
        the type of content you expect to find in the service response. The method will
        try to parse this content type if no content-type header is in the service's response, 
        otherwise the content-type in the response will be considered. So, it's a monkey-patch
        to deal with services written by idiots.

        Returns:
            service response (str or dictionary): Web service response, either a dictionary or a string.
    
    '''
    # prepare the URL
    if len(path_params)>0:
        url = url.format(**path_params)
    # prepare the payload
    if method in ['post', 'POST', 'put', 'PUT']:
        payload = prepare_payload(data, payload_content_type)
    else:
        payload = None
    # assemble the headers
    if additional_headers is not None:
        # add the extra headers provided as a dictionary
        headers = {**headers, **additional_headers}
    response = None
    # invoke the Web Service
    if method in ['post', 'POST']:
        response = requests.post(url, params = params, data=payload, headers=headers, files = files)
    elif method in ['put', 'PUT']:
        response = requests.put(url, params = params, data=payload, headers=headers, files=files)
    elif method in ['delete', 'DELETE']:
        response = requests.delete(url, params = params, data=payload, headers=headers)
    else:
        response=requests.get(url, params = params, data=payload, headers=headers, files = files)
    if response.status_code in [200,201,202,203,204,205]: # the 20X status means 'success'
        return parse_response(response, expected_response_content_type)
    else:
        logging.error('Request to '+str(url) + ' failed with code ' + str(response.status_code))
        response.raise_for_status() # Raises an exception for 4xx and 5xx status codes
    return None


def dataframe_records_gen(df):
    '''
	Creates a *generator* object allowing us to read dataframe rows as dictionaries.
	We need to do this because the bult-in to_dic function returns a friggin' list with
	a list comprehension, which is likely to clog your memory if the dataframe is large
	'''
    columns = df.columns.tolist()
    into_c = standardize_mapping(dict)
    for row in df.itertuples(index=False, name=None):
        yield into_c(
            (k, maybe_box_native(v)) for k, v in dict(zip(columns, row)).items()
        )
		
def dataframe_records_gen_fast(df):
    '''
    This is like dataframe_records_gen, but waaaay faster and with a smaller memory footprint, BUT
	it's rather unsafe, as it does not check whether or not row items are Scalars and/or composite objects
	before trying to shove them into the dict. If your df has nothing fancy in it, this should not be a 
	dealbreaker, but if you do have some weird stuff in there, oh boy... You are in for a world of pain'''
    cols = list(df)
    col_arr_map = {col: df[col].astype(object).to_numpy() for col in cols}
    for i in range(len(df)):
        yield {col: col_arr_map[col][i] for col in cols}


from setuptools import setup

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
   name='envrihub',
   version='0.0.9',
   description='Python library to streamline interaction with the ENVRI-Hub APIs, providing a pythonic facade to data and service access.',
   license="MIT",
   long_description=long_description,
   long_description_content_type="text/markdown",
   author='ENVRI Community',
   author_email='envri-hub-next-wp13-14@mailman.egi.eu',
   url="https://gitlab.a.incd.pt/envri-hub-next/vre-lib",
   packages=['envrihub', 'envrihub.cos', 'envrihub.data_access'],
   install_requires=['shapely','retry','openapi3-parser','prance','pandas','argopy','SPARQLWrapper'], #external packages as dependencies
)


from abc import ABC, abstractmethod
from datetime import datetime
import logging
import requests

from .models import Distribution, DistributionError
from .utils import datetime_to_epos_format

class Catalogue(ABC):
    '''The interface all catalogue-like objects are expected to implement'''
    @abstractmethod
    def search(self, query=None, **kwargs):
        pass


class EposCatalogue(Catalogue):
    '''this class encapsulates interactions with the catalogue of services
    check the Swagger documentation for advanced usage
    https://catalogue.staging.envri.eu/api/v1/ui
    '''
    base_address = 'https://catalogue.staging.envri.eu/api/v1'
    search_path = '/resources/search'
    details_path = '/resources/details/'
    
    def __init__(self, api_url = None):
        if api_url is None:
            self.api_url = self.base_address
        else:
            self.api_url = api_url
            
    def __parse_distributions(self, dists, facet = None):
        '''parses distribution JSON into a Distribution object'''
        for d in dists:
            try:
                yield Distribution.from_dict(d)
            except DistributionError as e:
                logging.error(e)


    def search(self, query:str='', startDate:datetime=None, endDate:datetime=None, bbox:tuple =None, keywords:list[str] =None,
               sciencedomains = None, servicetypes= None, organisations=None, exvs=None, facetstype = None,
               facets:bool = False):
        '''
        Allows to retrieve datasets as Distribution objects.
        You can query teh cataloge with a full text query and/or in a faceted way
        including time boundaries, geographical baoundaries, and other types of filter.

        Parameters
        ----------
        query:str
            A textual query in the likes of what you'd insert into a search engine
        startDate:datetime
            the beginning of the time interval of interest
        endDate: datetime
            the end of the time interval of interest
        bbox: tuple
            the four margins of the geographical area of interest.
            e.g. (10.70, 36.17, 28.98, 48.34)
        keywords: list[str]
            a list of keywords to filter research
        sciencedomains:

        servicetypes

        organisations

        exvs:str

        facestype: str
            one of the following values: 'categories', 'dataproviders', or 'serviceproviders'

        facets:bool
            whether or not to include facets in the search, which changes the structure of the
            output returned by the catalogue

        '''
        search_endpoint = self.api_url + self.search_path
        params = {
            'q': query, 
            'startDate': datetime_to_epos_format(startDate) if isinstance(startDate, datetime) else startDate,
            'endDate': datetime_to_epos_format(endDate) if isinstance(endDate, datetime) else endDate,
            'bbox': str(bbox)if bbox is not None else None,
            'keywords':str(keywords) if keywords is not None else None,
            'sciencedomains': sciencedomains,
            'servicetypes':servicetypes,
            'organisations':organisations,
            'exvs':str(exvs) if exvs is not None else None,
            'facetstype': facetstype if facetstype in ['categories', 'dataproviders', 'serviceproviders'] else None,
            'facets': facets
            
        }
        response_search = requests.get(search_endpoint, params=params)
        data = response_search.json()
        # parsing the JSON response to find distribution items
        if data.get('results') is not None:
            results = data.get('results')
            # if it's a faceted search, distributions are split into several
            # classes that are represented with a nested structure that looks like this
            # 'results' -> {'children':[
            #                    {'children':[
            #                          {'distributions':[...], 
            #                           'name':facet name},
            #                            ...
            #                                ]
            #                           },
            #                           ...
            #                       ]
            #              }
            if facets:
                if isinstance(results.get('children'), list):
                    for child in results.get('children'):
                        if isinstance(child.get('children'), list):
                            for nephew in child.get('children'):
                                if nephew.get('distributions') is not None:
                                     yield from self.__parse_distributions(nephew.get('distributions'), facet = nephew.get('name'))    
                                
            # else it's not a faceted search
            elif results.get('distributions') is not None:
                yield from self.__parse_distributions(results.get('distributions'))
        else:
            return None
                            


    def retrieve_resource(self, id):
        retrieve_endpoint = self.api_url + self.details_path
        response_retrieve = requests.get(retrieve_endpoint + '/' + str(id))
        return Distribution.from_dict(response_retrieve.json())
    
    